package be.iminds.ilabt.jfed.highlevel.jobs.states;

import be.iminds.ilabt.jfed.espec.ESpecLogic;
import be.iminds.ilabt.jfed.espec.bundle.ESpecBundle;
import be.iminds.ilabt.jfed.espec.model.ExecuteSpec;
import be.iminds.ilabt.jfed.espec.model.ExperimentSpecification;
import be.iminds.ilabt.jfed.espec.model.UploadLikeWithLogSpec;
import be.iminds.ilabt.jfed.espec.util.ESpecLogger;
import be.iminds.ilabt.jfed.espec.util.LimitedLiveLog;
import be.iminds.ilabt.jfed.espec.util.LimitedLiveLogImpl;
import be.iminds.ilabt.jfed.experiment.Experiment;
import be.iminds.ilabt.jfed.experiment.tasks.ExperimentTaskStatus;
import be.iminds.ilabt.jfed.highlevel.jobs.*;
import be.iminds.ilabt.jfed.highlevel.util.ExternalFileUtil;
import be.iminds.ilabt.jfed.highlevel.util.InputStreamToLogsThread;
import be.iminds.ilabt.jfed.highlevel.util.LogEntryGeneratorWrappingLogger;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedException;
import be.iminds.ilabt.jfed.rspec.model.ModelRspecType;
import be.iminds.ilabt.jfed.rspec_fx.model.javafx_impl.FXModelRspec;
import be.iminds.ilabt.jfed.rspec_fx.model.javafx_impl.FXRspecNode;
import be.iminds.ilabt.jfed.rspec.rspec_source.RequestRspecSource;
import net.schmizz.sshj.SSHClient;
import net.schmizz.sshj.connection.channel.direct.Session;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.slf4j.event.Level;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Objects;

/**
 * ExperimentSpecificationServicesDirsAndUploadsState executes a
 * single execute statement in an ExperimentSpecification, on all nodes (that are specified).
 * It waits until the execute is completed before being finished
 */
public class ExperimentSpecificationServicesSingleExecuteState extends SlicedState<ExperimentSpecificationServicesSingleExecuteState.ExperimentSpecificationServicesSingleExecuteStateSlice> {
    private static final Logger ACTUAL_LOG = LoggerFactory.getLogger(ExperimentSpecificationServicesSingleExecuteState.class);
    private final Logger LOG;
    public static final int MAX_EXECUTE_FINISHED_FILE_DETECTION_TRIES = 60;
    public static final int TIME_BETWEEN_FINISHED_FILE_DETECTION_TRIES_MS = 10_000;

    @Nonnull private final SetupSoftwareExperimentJob job;
    @Nonnull private final Experiment experiment;

    @Nonnull private final ExperimentSpecification experimentSpecification;
    @Nonnull private final ESpecBundle experimentSpecificationBundle;
    @Nonnull private final ESpecLogger eSpecLogger;

    @Nonnull private final List<ExperimentSpecificationServicesSingleExecuteStateSlice> slices;

    @Nonnull private final RemotePathResolver remotePathResolver;
    @Nonnull private final ExecuteSpec executeSpec;


    protected ExperimentSpecificationServicesSingleExecuteState(@Nonnull SetupSoftwareExperimentJob job,
                                                                @Nonnull RemotePathResolver remotePathResolver,
                                                                @Nonnull ExecuteSpec executeSpec,
                                                                boolean isInAnsible) {
        super("ExperimentSpecification execute "+ executeSpec.getDesc());
        this.job = job;
        this.LOG = new LogEntryGeneratorWrappingLogger(ACTUAL_LOG, job.getJobReport());
        this.experiment = job.getExperiment();
        this.eSpecLogger = job.getExperiment().getExperimentSpecificationLogger();
        this.remotePathResolver = remotePathResolver;
        this.executeSpec = executeSpec;

        if (experiment.getExperimentSpecification() == null)
            throw new RuntimeException("Internal error, ExperimentSpecificationServicesDirsAndUploadsState created for experiment without ExperimentSpecification");
        if (experiment.getExperimentSpecificationBundle() == null)
            throw new RuntimeException("Internal error, ExperimentSpecificationServicesDirsAndUploadsState created for experiment without ExperimentSpecification (bundle)");

        this.experimentSpecification = experiment.getExperimentSpecification();
        this.experimentSpecificationBundle = experiment.getExperimentSpecificationBundle();

        experiment.requireSlice();

        this.slices = new ArrayList<>();

        assert !isInAnsible || experimentSpecification.getAnsible() != null;

        RequestRspecSource requestRspecSource = experiment.getSlice().getRequestRspec();
        FXModelRspec model = requestRspecSource == null ?
                null :
                requestRspecSource.getModelRspec(ModelRspecType.FX);
        assert model != null;

        if (model == null) {
            LOG.warn("ExperimentSpecificationServicesSingleExecuteState will do nothing because there is no usable request RSpec");
            return;
        }

        for (FXRspecNode node : model.getNodes()) {
            if (Objects.equals(node.getSliverTypeName(), "fake"))
                continue;

            String nodeId = node.getClientId();

            if (nodeId != null) {
                if (ESpecLogic.mustRunOn(
                        nodeId, node.getAnsibleGroups(),
                        experimentSpecification, executeSpec)) {
                    slices.add(new ExperimentSpecificationServicesSingleExecuteStateSlice(job, node));
                }
            }
        }
    }

    @Nonnull
    @Override
    protected ExperimentTaskStatus executeState(Job<?> job) throws InterruptedException, JFedException {
        ExperimentTaskStatus res = super.executeState(job);
        experiment.getExperimentSpecificationLogger().firePostExecuteStepAll(executeSpec, res == ExperimentTaskStatus.SUCCESS);
        return res;
    }

    @Override
    public Collection<ExperimentSpecificationServicesSingleExecuteStateSlice> getSlices() {
        return slices;
    }

    public class ExperimentSpecificationServicesSingleExecuteStateSlice extends StateSlice {
        @Nonnull protected final SetupSoftwareExperimentJob setupSoftwareExperimentJob;
        @Nonnull protected final FXRspecNode node;
        @Nonnull protected final String clientId;

        public ExperimentSpecificationServicesSingleExecuteStateSlice(@Nonnull SetupSoftwareExperimentJob job,
                                                                      @Nonnull FXRspecNode node) {
            super(job);
            this.setupSoftwareExperimentJob = job;
            this.node = node;

            if (node.getClientId() == null)
                throw new IllegalStateException("Only nodes with a client_id can be used in an ExperimentSpecification");
            this.clientId = node.getClientId();
        }

        @Override
        public String getName() {
            return "ExperimentSpecification execute steps for node "+node.getClientId();
        }

        @Override
        public ExperimentTaskStatus statefulRun() throws JFedException, InterruptedException {
            if (Objects.equals(node.getSliverTypeName(), "fake")) {
                LOG.warn("execute should NOT run on node with sliver type 'fake' clientId="+clientId+"  " +
                        "(this is a bug in the code, and it will be handled by not running the execute statefulRun)");
                return ExperimentTaskStatus.SUCCESS;
            }

            SSHClient ssh;
            try {
                ssh = setupSoftwareExperimentJob.getSSHClient(node, this);
            } catch (SetupSoftwareExperimentJob.SshException e) {
                throw new IllegalStateException("There is no SSH connection for node \""+clientId+"\" available", e);
            }

            if (!ssh.isConnected()) {
                updateMessage("SSH connection is not connection: will abort execute.");
                return ExperimentTaskStatus.FAILED;
            }

            final ExecuteState executeState = new ExecuteState(ssh);
            setAndRunState(executeState);

            if (executeState.getStatus() != ExperimentTaskStatus.SUCCESS)
                return ExperimentTaskStatus.FAILED;
//
//            //TODO: later, we might need something like this for ADVANCED scripts
////            //check if there are finished-files to wait for
////            if (node.getExecuteServices().stream().anyMatch(es -> es.getFinishedFile() != null)) {
////                setAndRunState(new WaitForExecuteServicesState());
////            } else {
////                LOG.info("Did not find any execute-services with finished-file attribute. We assume that all files that are needed are present!");
////            }

            return ExperimentTaskStatus.SUCCESS;
        }

//        public class WaitForExecuteServicesState extends State {
//
//            protected WaitForExecuteServicesState() {
//                super("Waiting for execute-services to finish on " + node.getClientId());
//            }
//
//            @Nonnull
//            @Override
//            protected ExperimentTaskStatus executeState(Job<?> job) throws InterruptedException, JFedException {
//                try {
//                    for (FXExecuteService executeService : node.getExecuteServices()) {
//                        if (executeService.getFinishedFile() == null)
//                            continue;
//
//                        boolean fileFound = false;
//                        int tries = 0;
//
//                        while (!fileFound && tries < MAX_EXECUTE_FINISHED_FILE_DETECTION_TRIES) {
//                            Session.Command command =
//                                    ssh.startSession().exec("test -f " + executeService.getFinishedFile());
//                            command.join();
//
//                            fileFound = command.getExitStatus() == 0;
//
//                            tries++;
//                            if (!fileFound) {
//                                updateMessage("Looked for '" + executeService.getFinishedFile() + "' " + tries + " times.");
//                                Thread.sleep(TIME_BETWEEN_FINISHED_FILE_DETECTION_TRIES_MS);
//                            }
//                        }
//
//                        if (!fileFound) {
//                            updateMessage("Timed out while waiting for finished-file '" + executeService.getFinishedFile() + "' to appear");
//                            return ExperimentTaskStatus.FAILED;
//                        } else {
//                            updateMessage("Found finished-file '" + executeService.getFinishedFile() + "'");
//                        }
//                    }
//
//                    updateMessage("Found all finished-files");
//                    return ExperimentTaskStatus.SUCCESS;
//
//
//                } catch (IOException ex) {
//                    LOG.error("Error while waiting for finished-files on {}", node.getClientId(), ex);
//                    updateMessage("Error while waiting for finished-files on {}" + ex.getMessage());
//                    return ExperimentTaskStatus.FAILED;
//                }
//            }
//        }

        public class ExecuteState extends State {
            @Nonnull private final SSHClient ssh;

            protected ExecuteState(@Nonnull SSHClient ssh) {
                super("Executing "+executeSpec.getDesc()+" on " + node.getClientId());
                this.ssh = ssh;
            }

            @Nonnull
            @Override
            protected ExperimentTaskStatus executeState(Job<?> job) throws InterruptedException, JFedException {
                if (ssh == null)
                    throw new IllegalStateException("ssh == null");

                String desc = executeSpec.getDesc();
                long logId = eSpecLogger.getNextLogId();

                if (!ESpecLogic.mustRunOn(
                        clientId, node.getAnsibleGroups(),
                        experimentSpecification, executeSpec)) {
                    throw new IllegalStateException("BUG: execute "+desc+" does not need to be executed on this " +
                            "node (\""+clientId+"\") (nodes="+executeSpec.getNodes()+")");
                }

                String remotePath = remotePathResolver.getRemotePath(executeSpec, clientId);
                if (remotePath == null) {
                    eSpecLogger.firePostExecuteFail(logId, executeSpec, "not found", null, clientId, -1,
                            "Did not find remote path of executable. Bug in execute step?", null);
                    LOG.error("Did not find remote path of executable. Bug in execute step?");
                    updateMessage("Did not find remote path of executable. Bug in execute step?");
                    throw new RuntimeException("Did not find remote path of executable. Bug in execute step?");
//                    return ExperimentTaskStatus.FAILED;
                }

                String remoteLogPath = remotePathResolver.getRemoteLogPath(executeSpec, clientId);
                SshCommandAndLog commandAndLog = createSshCommand(remotePath, remoteLogPath,
                        executeSpec, experimentSpecification, clientId,
                        executeSpec.getPwd(),
                        executeSpec.isSudo());

                try {
                    LOG.debug("ExperimentSpecification.execute is executing {} on {}",
                            remotePath, node.getClientId());
                    LOG.debug("      command: {}", commandAndLog.getCommand());


                    LimitedLiveLog log = new LimitedLiveLogImpl(1_000_000);
                    eSpecLogger.firePreExecute(logId, executeSpec, remotePath,
                            commandAndLog.getLogPath(), clientId, log);

                    Session.Command c =
                            ssh.startSession().exec(commandAndLog.getCommand());

                    new InputStreamToLogsThread(c.getInputStream(), LOG, Level.TRACE,// Level.INFO,
                            desc+"-OUTPUT+"+node.getClientId()+": ", log).start();
                    new InputStreamToLogsThread(c.getErrorStream(), LOG, Level.DEBUG, //Level.WARN,
                            desc+"-ERROR"+node.getClientId()+": ", log).start();

                    c.join();

                    int exitStatus = c.getExitStatus();

                    if (exitStatus != 0) {
                        eSpecLogger.firePostExecuteFail(logId, executeSpec, remotePath, commandAndLog.getLogPath(), clientId, exitStatus, desc+
                                " execution was unsuccessful. Aborting", null);
                        updateMessage(desc+" execution was unsuccessful. Aborting");
                        LOG.error("ExperimentSpecification.execute {} on {} FAILURE: " +
                                        "exitstatus = {}", remotePath, node.getClientId(), exitStatus);
                        return ExperimentTaskStatus.FAILED;
                    } else {
                        LOG.debug("ExperimentSpecification.execute {} on {} success: " +
                                        "exitstatus = {}", remotePath, node.getClientId(), exitStatus);
                        eSpecLogger.firePostExecuteSuccess(logId, executeSpec, remotePath, commandAndLog.getLogPath(), clientId, exitStatus);
                        return ExperimentTaskStatus.SUCCESS;
                    }
                } catch (IOException ex) {
                    eSpecLogger.firePostExecuteFail(logId, executeSpec, remotePath, commandAndLog.getLogPath(), clientId, -1,
                            "IOException while executing "+remotePath+" on "+node.getClientId(), ex);
                    LOG.error(String.format("IOException while executing %s on %s", remotePath, node.getClientId()), ex);
                    updateMessage(String.format("Error while executing %s on %s: %s", remotePath, node.getClientId(), ex.getMessage()));
                    return ExperimentTaskStatus.FAILED;
                }
            }
        }
    }

    private static class SshCommandAndLog {
        @Nonnull private final String command;
        @Nullable private final String logPath;

        public SshCommandAndLog(@Nonnull String command, @Nullable String logPath) {
            this.command = command;
            this.logPath = logPath;
        }

        @Nonnull
        public String getCommand() {
            return command;
        }

        @Nullable
        public String getLogPath() {
            return logPath;
        }
    }
    public static SshCommandAndLog createSshCommand(@Nonnull String remotePath,
                                                    @Nullable String remoteLogPath,
                                                    @Nonnull UploadLikeWithLogSpec executeSpec,
                                                    @Nonnull ExperimentSpecification experimentSpecification,
                                                    @Nonnull String clientId,
                                                    @Nullable String pwd,
                                                    @Nullable Boolean useSudo) throws JFedException, InterruptedException {
        if (pwd == null) {
            //use working dir of script
            pwd = ExternalFileUtil.getPathOnly(remotePath);
        }

        return new SshCommandAndLog(
                createSshCommand(remotePath, pwd, remoteLogPath, useSudo, experimentSpecification),
                remoteLogPath
        );
    }

    public static String createSshCommand(@Nonnull String baseCommand,
                                          @Nullable String pwd,
                                          @Nullable String logPath,
                                          @Nullable Boolean useSudo,
                                          @Nonnull ExperimentSpecification experimentSpecification) {
        boolean localLogs = experimentSpecification.getConfig().isStoreLocalLogsByDefault();

        String sudoargs = "";
        if (experimentSpecification.getConfig().getDefaultSudoUser() != null) {
            sudoargs += "-u "+experimentSpecification.getConfig().getDefaultSudoUser()+" ";
        }

        if (useSudo == null) {
            useSudo = experimentSpecification.getConfig().isUseSudoByDefault();
        }

        String hereDocCommand = "<<'EOF'";

        String command;
        //old, which only worked when remote shell was bash, but failed on the pipefail due to dash
//        if (useSudo) {
//            if (logPath != null) {
//                if (localLogs)
//                    command = "set -o pipefail; sudo " + sudoargs + "bash " + hereDocCommand + " 2>&1 | sudo tee " + logPath + "";
//                else
//                    command = "sudo " + sudoargs+"bash " + hereDocCommand + " > " + logPath + " 2>&1'";
//            } else {
//                if (localLogs)
//                    command = "sudo " + sudoargs + hereDocCommand;
//                else
//                    command = "sudo " + sudoargs + hereDocCommand+" > /dev/null 2>&1";
//            }
//        } else {
//            if (logPath != null) {
//                if (localLogs)
//                    command = "set -o pipefail'; bash "+ hereDocCommand + " 2>&1 | tee '" + logPath + "'";
//                else
//                    command = "bash "+hereDocCommand + " > '" + logPath + "' 2>&1";
//            } else {
//                if (localLogs) {
////                    return "bash " + baseCommand;
//                    command = "bash "+hereDocCommand+"";
//                } else
//                    command = "bash "+hereDocCommand+" > /dev/null 2>&1";
//            }
//        }
        //new
        if (useSudo) {
            command = "sudo " + sudoargs + " -- ";
        } else {
            command = "";
        }
        if (localLogs) {
            //later we handle if we also need to log to file
            command += "bash "+hereDocCommand+"";
        } else {
            if (logPath != null) {
                //log only to file
                command += "bash " + hereDocCommand + " >'" + logPath + "' 2>'" + logPath + "'";
            } else {
                //no logging needed at all
                command += "bash " + hereDocCommand + " > /dev/null 2>&1";
            }
        }

        command += "\n";
        if (pwd != null) {
            command += "cd '"+pwd+"'\n";
        }
        if (logPath != null && localLogs) {
            //log to both stdin/stdout AND to file   (in current version, both stdout and stdin are sent to stdout)
//            command += "exec 3>&1 4>&2 1> >(tee '" + logPath + "' >&3) 2> >(tee '" + logPath + "' >&4)\n";
            command += "exec 3>&1 4>&2 5> >(tee '" + logPath + "' >&3) 1>&5 2>&5\n";
        }
        command += baseCommand+"\n";
        command += "EOF\n";
        return command;
    }

}
